"""
Copyright 2008-2014 by Aerospike, Inc. All rights reserved.
THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE. THE COPYRIGHT NOTICE
ABOVE DOES NOT EVIDENCE ANY ACTUAL OR INTENDED PUBLICATION.
"""


# -*- coding: utf-8 -*-
from libs import base
from libs import management
from constants import *
from models import namespace
from models import xdr
from alerts import node_stats
from helpers.info_dict import InfoDict
from collections import deque
from datetime import datetime
import calendar
import time


class Node(object):
    """Node contains all information about node"""

    def __new__(cls, address, cluster_id, validity=True, user=None, password=None):

        ip, port = address.split(":")
        #if base.is_node_available(ip, port) is False:
        #    return None
        return super(Node, cls).__new__(cls, address, cluster_id, validity,  user=user, password=password)

    def __init__(self, address, cluster_id, validity=True, user=None, password=None):

        ip, port = address.split(":")

        self._ip = ip
        self._port = port
        self._address = "%s:%s" % (ip, port)
        self._cluster_id = cluster_id

        self.user = user
        self.password = password

        self._status = "on"
        if base.is_node_available(ip, port, user=user, password=password) is False:
            self._status = "off"

        self._cluster_visibility = True

        self._id = None
        self._latency = None
        self._bins_info = None
        self._statistics = {}
        self._modified_statistics = {}
        self._configuration = {}
        self._modified_configuration = {}
        self._validity = validity

        self._namespace_list = set()
        self._namespaces = {}
        self._xdr = None
        self._build = base.get_node_build(self._ip, self._port, user=user, password=password)

        timestamp = datetime.utcnow()
        timestamp = int(calendar.timegm(timestamp.timetuple(
            )) * 1e3 + timestamp.microsecond / 1e3)
        self._read_tps_history = deque([dict(x=timestamp, secondary=None, y=None)], maxlen=TPS_HISTORY_LIMIT)
        self._write_tps_history = deque([dict(x=timestamp, secondary=None, y=None)], maxlen=TPS_HISTORY_LIMIT)

        base.configure_node_latency(self._ip, self._port, user=user, password=password)
        latency = base.get_node_latency(self._ip, self._port, user=user, password=password)
        self._latency_history = deque([latency], maxlen=LATENCY_HISTORY_LIMIT)

        self._stats_alert = node_stats.NodeStats(self._cluster_id, self._address, self._status)

    def get_address(self):

        return self._address

    def get_status(self, user=None, password=None):

        return self._status

    def set_status(self, status):

        self._status = status

    def get_cluster_visibility(self):

        return self._cluster_visibility

    def set_cluster_visibility(self, cluster_visibility):

        self._cluster_visibility = cluster_visibility

    def get_id(self, user=None, password=None):

        if self._id is None:
            self._id = base.get_node_id(self._ip, self._port,
                                        user=user, password=password)
        return self._id

    def get_build(self):

        return self._build

    def update_build(self, user=None, password=None):

        self._build = base.get_node_build(self._ip, self._port, user=user, password=password)
        return self._build

    def get_cluster_key(self, user=None, password=None):

        statistics = base.get_node_statistics(self._ip, self._port, user=user, password=password)
        return statistics.get("cluster_key", UNABLE_TO_CONNECT)

    def get_statistic(self, attr):

        return self._statistics.get(attr, NOT_SUPPORTED)

    def get_statistics(self):

        return self._statistics

    def get_modified_statistics(self):

        return self._modified_statistics

    def get_read_tps_history(self, start_time, end_time):

        if start_time is not None and end_time is not None:
            return filter(lambda tps: tps['x'] > start_time and tps['x'] < end_time, list(self._read_tps_history))
        elif start_time is not None:
            return filter(lambda tps: tps['x'] > start_time, list(self._read_tps_history))
        elif end_time is not None:
            return filter(lambda tps: tps['x'] < end_time, list(self._read_tps_history))
        return self._read_tps_history

    def get_write_tps_history(self, start_time, end_time):

        if start_time is not None and end_time is not None:
            return filter(lambda tps: tps['x'] > start_time and tps['x'] < end_time, list(self._write_tps_history))
        elif start_time is not None:
            return filter(lambda tps: tps['x'] > start_time, list(self._write_tps_history))
        elif end_time is not None:
            return filter(lambda tps: tps['x'] < end_time, list(self._write_tps_history))
        return self._write_tps_history

    def get_read_tps(self):

        return self._read_tps_history[-1]

    def get_write_tps(self):

        return self._write_tps_history[-1]

    def get_latency_history(self, start_time, end_time):

        if start_time is not None and end_time is not None:
            start_time = time.strptime(start_time, "%H:%M:%S")
            end_time = time.strptime(end_time, "%H:%M:%S")
            return filter(lambda x: time.strptime(x.values()[0].get('timestamp'), "%H:%M:%S") > start_time and time.strptime(x.values()[0]['timestamp'], "%H:%M:%S") < end_time if x != {} else False, list(self._latency_history))
        elif start_time is not None:
            start_time = time.strptime(start_time, "%H:%M:%S")
            return filter(lambda x: time.strptime(x.values()[0].get('timestamp'), "%H:%M:%S") > start_time if x != {} else False, list(self._latency_history))
        elif end_time is not None:
            end_time = time.strptime(end_time, "%H:%M:%S")
            return filter(lambda x: time.strptime(x.values()[0].get('timestamp'), "%H:%M:%S") < end_time if x != {} else False, list(self._latency_history))
        return self._latency_history

    def get_latency(self):

        return self._latency_history[-1]

    def set_invalid(self):

        self._validity = False

    def set_valid(self):

        self._validity = True

    def set_unknown(self):

        self._validity = None

    def is_valid(self):

        return self._validity

    def get_updated_latencies(self, user=None, password=None):

        latency = base.get_node_latency(self._ip, self._port, user=user, password=password)
        self._latency_history.append(latency)

    def get_updated_statistics(self, cluster, user=None, password=None):

        old_stat_write_reqs = self.get_statistic("stat_write_reqs")
        old_stat_write_success = self.get_statistic("stat_write_success")
        old_stat_read_reqs = self.get_statistic("stat_read_reqs")
        old_stat_read_success = self.get_statistic("stat_read_success")

        statistics = base.get_node_statistics(self._ip, self._port, user=user, password=password)
        if statistics.get("node_status") == "off":
            self.set_unknown()
        configuration = self.get_updated_configurations(user=user, password=password)
        self._modified_configuration = dict(set(configuration.items()) - set(self._configuration.items()))
        self._configuration = configuration
        node_cluster_key = statistics.get("cluster_key", None)
        if node_cluster_key == cluster.get_cluster_key(user=user, password=password):
            self.set_valid()
        elif self.is_valid() is None and node_cluster_key != None and node_cluster_key != cluster.get_cluster_key(user=user, password=password):
            self.set_invalid()
            return
        self._modified_statistics = dict(set(statistics.items()) - set(self._statistics.items()))
        self._statistics = statistics

        timestamp = datetime.utcnow()
        timestamp = int(calendar.timegm(timestamp.timetuple())*1e3 + timestamp.microsecond/1e3)

        new_stat_write_reqs = self.get_statistic("stat_write_reqs")
        new_stat_write_success = self.get_statistic("stat_write_success")
        new_stat_read_reqs = self.get_statistic("stat_read_reqs")
        new_stat_read_success = self.get_statistic("stat_read_success")

        if True not in map(lambda x: x == NOT_SUPPORTED, [new_stat_write_reqs, old_stat_write_reqs, new_stat_write_success, old_stat_write_success]):
            total_write_tps = int(new_stat_write_reqs) - int(old_stat_write_reqs)
            success_write_tps = int(new_stat_write_success) - int(old_stat_write_success)
            old_timestamp = self._write_tps_history[-1]['x']
            time_difference = datetime.fromtimestamp((timestamp - old_timestamp)/1000.0).second + (datetime.fromtimestamp((timestamp - old_timestamp)/1000.0).microsecond)/1000000.0
            total_write_tps = total_write_tps / time_difference
            total_write_tps = round(total_write_tps, 1) if total_write_tps < 1 else round(total_write_tps)
            success_write_tps = success_write_tps / time_difference
            success_write_tps = round(success_write_tps, 1) if success_write_tps < 1 else round(success_write_tps)
            self._write_tps_history.append(dict(x=timestamp, secondary=total_write_tps, y=success_write_tps))
        else:
            self._write_tps_history.append(dict(x=timestamp, secondary=None, y=None))

        if True not in map(lambda x: x == NOT_SUPPORTED, [new_stat_read_reqs, old_stat_read_reqs, new_stat_read_success, old_stat_read_success]):
            total_read_tps = int(new_stat_read_reqs) - int(old_stat_read_reqs)
            success_read_tps = int(new_stat_read_success) - int(old_stat_read_success)
            old_timestamp = self._read_tps_history[-1]['x']
            time_difference = datetime.fromtimestamp((timestamp - old_timestamp)/1000.0).second + (datetime.fromtimestamp((timestamp - old_timestamp)/1000.0).microsecond)/1000000.0
            total_read_tps = total_read_tps / time_difference
            total_read_tps = round(total_read_tps, 1) if total_read_tps < 1 else round(total_read_tps)
            success_read_tps = success_read_tps / time_difference
            success_read_tps = round(success_read_tps, 1) if success_read_tps < 1 else round(success_read_tps)
            self._read_tps_history.append(dict(x=timestamp, secondary=total_read_tps, y=success_read_tps))
        else:
            self._read_tps_history.append(dict(x=timestamp, secondary=None, y=None))
        self._statistics['cluster_visibility'] = self.get_cluster_visibility()

        ##    * Unit test cases *
        ##from random import randrange

        ##free_pct_disk = ["2", "5", "10"]
        ##self._statistics['free-pct-disk'] = free_pct_disk[randrange(0, 3)]

        ##client_connections = ["90", "95", "96"]
        ##self._statistics['client_connections'] = client_connections[randrange(0, 3)]   def get_updated_statistics(self):


       ##client_connections = ["90", "95", "96"]
        ##self._statistics['client_connections'] = client_connections[randrange(0, 3)]
        ##self._statistics['proto-fd-max'] = "100"

        ##queue = ["10001", "9999"]
        ##self._statistics['queue'] = queue[randrange(0, 2)]

        ##cluster_visibility = [True, False]
        ##self._statistics['cluster_visibility'] = cluster_visibility[randrange(0, 2)]

        self.update_build(user=user, password=password)
        return self._statistics

    def get_subset_of_statistics(self, attrs=[]):

        subset = InfoDict()
        for attr in attrs:
            subset[attr] = self._statistics.get(attr, NOT_SUPPORTED)
        return subset

    def get_disk(self):

        return self.get_statistic("used-bytes-disk"), self.get_statistic("total-bytes-disk")

    def get_disk_stats(self):

        attrs = ["used-bytes-disk", "total-bytes-disk"]
        stats = self.get_subset_of_statistics(attrs)
        try:
            stats["used-bytes-disk"] = int(stats["used-bytes-disk"])
            stats["total-bytes-disk"] = int(stats["total-bytes-disk"])
            stats["free-bytes-disk"] = stats["total-bytes-disk"] - stats["used-bytes-disk"]
        except:
            stats["free-bytes-disk"] = NOT_SUPPORTED
        return stats

    def get_memory(self):

        return self.get_statistic("used-bytes-memory"), self.get_statistic("total-bytes-memory")

    def get_memory_stats(self):

        attrs = ["used-bytes-memory", "total-bytes-memory"]
        stats = self.get_subset_of_statistics(attrs)
        try:
            stats["used-bytes-memory"] = int(stats["used-bytes-memory"])
            stats["total-bytes-memory"] = int(stats["total-bytes-memory"])
            stats["free-bytes-memory"] = stats["total-bytes-memory"] - stats["used-bytes-memory"]
        except:
            stats["free-bytes-memory"] = NOT_SUPPORTED
        return stats

    def get_bins_info(self, user=None, password=None):

        self._bins_info = base.get_node_bins_info(self._ip, self._port,
                                    user=user, password=password)
        return self._bins_info

#    def get_latency(self):

#        self._latency = base.get_node_latency(self._ip, self._port)
#        return self._latency

    def get_throughput(self):

        throughput = InfoDict(read_tps=self.get_statistic("read_tps"),
            write_tps=self.get_statistic("write_tps"))
        return throughput

    def get_migration_stats(self):

        attrs = ["migrate_msgs_sent", "migrate_msgs_recv", "migrate_progress_send",
            "migrate_progress_recv", "migrate_tx_objs", "migrate_rx_objs"]
        return self.get_subset_of_statistics(attrs)

    def get_configuration(self, attr):

        return configurations.get(attr, NOT_SUPPORTED)

    def get_configurations(self):

        return self._configuration

    def get_modified_configurations(self):

        return self._modified_configuration

    def get_updated_configurations(self, user=None, password=None):

        configuration = base.get_node_configuration(self._ip, self._port,
                                        user=user, password=password)
        self._modified_configuration = dict(set(configuration.items()) - set(self._configuration.items()))
        self._configuration = configuration
        return self._configuration

    def update_configurations(self, user=None, password=None):

        configuration = base.get_node_configuration(self._ip, self._port,
                                clear_cache=True, user=user, password=password)
        self._modified_configuration = dict(set(configuration.items()) - set(self._configuration.items()))
        self._configuration = configuration

    def set_configuration(self, parameter, value, user=None, password=None):

        status = base.set_node_configuration(self._ip, self._port,
            str(parameter), str(value), user=user, password=password)
        return status

    def set_configurations(self, edit_parameters, user=None, password=None):

        error_list = list()
        for key, value in list(edit_parameters.items()):
            if self.set_configuration(key, value, user=user, password=password) is False:
                error_list.append(key)
        self.update_configurations(user=user, password=password)

        return error_list

    def get_subset_of_configurations(self, attrs=[]):

        subset = InfoDict()
        configurations = self.get_configurations()

        if configurations is None:
            return {}

        for attr in attrs:
            subset[attr] = configurations.get(attr, NOT_SUPPORTED)
        return subset

    def get_xdr(self, xdr_port, user=None, password=None):

        return xdr.Xdr(self._ip, self._port, xdr_port, self._cluster_id, user=user, password=user)
        #return self._xdr

    def get_namespaces(self):

        return self._namespaces

    def get_namespace(self, name):

        return self._namespaces.get(name, None)

    def get_node_list(self, user=None, password=None):

        if not self.is_valid():
            return []
        self._node_list = base.get_cluster_nodes(self._ip, self._port,
                                            user=user, password=password)
        return self._node_list

    def get_namespace_list(self, user=None, password=None):

        if not self.is_valid():
            return []
        namespace_list = base.get_namespaces(self._ip, self._port,
                                            user=user, password=password)
        if namespace_list == UNABLE_TO_CONNECT:
            return self._namespace_list

        new_namespaces = namespace_list - self._namespace_list
        self.add_namespaces(new_namespaces)

        deleted_namespaces = self._namespace_list - namespace_list
        self.delete_namespaces(deleted_namespaces)

        self._namespace_list = namespace_list
        return self._namespace_list

    def add_namespaces(self, namespace_names):

        for namespace_name in namespace_names:
            self.add_namespace(namespace_name)

    def add_namespace(self, namespace_name):

        self._namespaces[namespace_name] = namespace.Namespace(namespace_name, self._ip, self._port, self._cluster_id)

    def delete_namespaces(self, namespace_names):

        for namespace_name in namespace_names:
            self.delete_namespace(namespace_name)

    def delete_namespace(self, namespace_name):

        del self._namespaces[namespace_name]

    def update_alerts(self, user=None, password=None):
        stats = self.get_statistics()
        self._stats_alert.generate_alerts(stats, self.get_configurations(), self.get_status())
        for ns in list(self._namespaces.values()):
            ns.update_alerts(user=user, password=password)

    def get_scan_jobs(self, user=None, password=None):

        return base.get_scan_jobs(self._ip, self._port,
                                user=user, password=password)

    def get_jobs(self, user=None, password=None):

        try:
            return base.get_job_details(self._ip, self._port,
                                    user=user, password=password)
        except:
            return []

    def get_udfs(self, user=None, password=None):

        return base.get_udf_list(self._ip, self._port,
                                    user=user, password=password)

    def change_status(self, desired_status, username, password):

        if self.get_build() is None:
            response = management.change_node_status(desired_status, self._ip, username, password, "3")
            if response["status"] == "Failure":
                response = management.change_node_status(desired_status, self._ip, username, password, "2")
        else:
            response = management.change_node_status(desired_status, self._ip, username, password, self.get_build())
        if response["status"] == "Success" and desired_status == "on":
            self.set_status("on")
        elif response["status"] == "Success" and desired_status == "off":
            self.set_status("off")
        return response

    def change_xdr_status(self, desired_status, username, password):

        if self.get_build() is None:
            response = management.change_xdr_status(desired_status, self._ip, username, password, "3")
            if response["status"] == "Failure":
                response = management.change_xdr_status(desired_status, self._ip, username, password, "2")

        else:
            response = management.change_xdr_status(desired_status, self._ip, username, password, self.get_build())
        return response
